"""docstring."""


from difflib import get_close_matches

from sklearn.pipeline import Pipeline

from cleaner import TextProcessor
from lsi import GensimLsi
from vectorizers import GensimTfidf

import pickle

with open("org_data.pkl", 'rb') as fp:
    data = pickle.load(fp)
with open("name_list.pkl", 'rb') as fp:
    data_names = pickle.load(fp)

TFIDF = 'tfidf.pkl'
LSI = "test_lsi.pkl"
INDEX = "test_index.pkl"
ID2WORD = "tfidf_dict.pkl"


class OrgSim():
    """Find Org most similar to Article."""

    def __init__(self, lsi_path=LSI, id2word_path=ID2WORD, index_path=INDEX,
                 tfidf_path=TFIDF, org_data=data, name_list=data_names):
        """
        Initialize class.

        Parameters
        ----------
        lsi_path : str
            Path to location of saved gensim LsiModel.
            If specified, the model will load and use this object
            as its LsiModel.
        dict_path : str
            Path to location of saved gensim Dictionary.
            If specified, the model will load and use this object
            as its Dictionary.
        tfidf_path: str
            File-path designating where self.tfidf should be saved.
        org_data: list
            List of data.
        name_list: list
            List of names.
        """
        self.org_data = org_data
        self.name_list = name_list
        self.processor = TextProcessor()
        self.tfidf = GensimTfidf(tfidf_path=tfidf_path,
                                 dictionary_path=id2word_path,
                                 use_sparse_representation=True)
        self.lsi = GensimLsi(lsi_path=lsi_path,
                             id2word_path=id2word_path,
                             index_path=index_path)
        self.transformer = Pipeline([
            ('norm', self.processor),
            ('tfidf', self.tfidf)])

    @staticmethod
    def closest_match(string1, strings):
        """
        Return the most similar org in name_list.

        Parameters
        ----------
        string1: str
            String being queried.
        strings: list
            List of org names.
        """
        result = get_close_matches(string1, strings)
        try:
            return result[0]
        except IndexError:
            return "Not Found"

    def resolve_query(self, org):
        """
        Find most similar org to 'org'.

        Parameters
        ----------
        org: str
            Name of organizatin to query.
        """
        if org in set(self.name_list):
            correct_org = org
        else:
            correct_org = self.closest_match(org, self.name_list)
        # return associated data
        return correct_org, self.name_list.index(correct_org)

    def similarity(self, org, n=10):
        """
        Return the 10 most similar orgs to org.

        Parameters
        ----------
        org: string
            Name of org to query.
        n: int
            Number of orgs to return.

        Returns
        -------
            list: tuples of (org, similarity).
        """
        doc, idx = self.resolve_query(org)
        if doc == "Not Found":
            return "Org not found, please search for another name."
        # Find data associated with doc before returning anything
        doc_data = self.org_data[idx]
        tfidf_data = self.transformer.transform([doc_data])
        return self.lsi.similarity(doc=tfidf_data[0], n=n)


class Art2Org():
    """init."""

    pass


class Org2Art():
    """init."""

    pass



"""Lsi Model."""
import os
import pickle

from sklearn.base import BaseEstimator, TransformerMixin

with open("name_list.pkl", 'rb') as fp:
    data_names = pickle.load(fp)


# keep this
class GensimLsi(BaseEstimator, TransformerMixin):
    """
    Custom sklearn transformer to convert tokenized,
    preprocessed data to tf-idf representation.
    """
    def __init__(self, id2word_path, lsi_path=None, index_path=None):
        self.num_topics = 10
        self.corpus = None
        self.index = None
        self.model = None
        self.id2word = self.load(id2word_path=id2word_path)
        if lsi_path is not None:
            self.model = self.load(lsi_path=lsi_path)
        if index_path is not None:
            self.index = self.load(index_path=index_path)

    @staticmethod
    def load(lsi_path=None, id2word_path=None, index_path=None):
        """
        If specified, attempts to load gensim LsiModel from `lsi_path`
        and gensim Dictionary from `dictionary_path`.

        Parameters
        ----------
        lsi_path: str
            File-path designating where self.model should be saved.
        id2word_path: str
            File-path designating where self.dictionary should be saved.
        """
        if lsi_path is not None:
            from gensim.models import LsiModel
            if not os.path.exists(lsi_path):
                raise IOError('The provided file path to the LsiModel was not found.'
                              'Please ensure that the argument is the correct path.')
            return LsiModel.load(lsi_path)
        if id2word_path is not None:
            from gensim.corpora.dictionary import Dictionary
            if not os.path.exists(id2word_path):
                raise IOError('The provided file path to the Dictionary was not found.'
                              'Please ensure that the argument is the correct path.')
            return Dictionary.load(id2word_path)
        if index_path is not None:
            from gensim.similarities import MatrixSimilarity
            if not os.path.exists(index_path):
                raise IOError('The provided file path to the Dictionary was not found.'
                              'Please ensure that the argument is the correct path.')
            return MatrixSimilarity.load(index_path)

    def save(self, lsi_path=None, id2word_path=None, index_path=None):
        """
        Saves objects from fit process: gensim.LsiModel to `lsi_path`
        and gensim.Dictionary to `dictionary_path`.
        If either self.model or self.dictionary does not exist, an
        AttributeError is raised.

        Parameters
        ----------
        lsi_path: str
            File-path designating where self.model should be saved.
        dictionary_path: str
            File-path designating where self.dictionary should be saved.
        """
        if not (self.model and self.id2word):
            raise AttributeError('Nothing to save yet, please run .fit first.')
        if lsi_path is not None:
            self.model.save(lsi_path)
        if id2word_path is not None:
            self.id2word.save(id2word_path)
        if index_path is not None:
            self.index.save(index_path)


    def fit(self, documents, num_topics=600, labels=None):
        """
        Fits a gensim LsiModel to documents.

        Parameters
        ----------
        documents: iterable
            List of documents. Each document must be a list of preprocessed tokens.
        labels: iterable
            Optional list of same size as documents, specifying label for each document.

        """
        from gensim.models import LsiModel
        from gensim.corpora.dictionary import Dictionary
        if self.id2word is None:
            raise AttributeError("Must specify id2word.")
        self.num_topics = num_topics
        self.model = LsiModel(documents, id2word=self.id2word, num_topics=600)
        self.corpus = self.model[documents]
        return self

    def transform(self, documents):
        """
        Returns a vectorized embedding of each document in documents.

        Parameters
        -----------
        documents: iterable
            List of documents. Each document must be a list of tokens.

        Returns
        -------
            iterable: list of vectorized documents.
        """
        if self.model is None:
            raise AttributeError('Must have a fit model in order'
                                 ' to call transform.')
        return self.model[documents]

    def similarity(self, doc, n=10):
        """
        Returns the `n` most similar items in `self.corpus`
        to `doc`.

        Parameters
        ----------
        doc:
            A document. embedded in same tfidf space as model.
        n: int (default=10)
            Number of most similar items to return.

        Returns
        -------
            sims: dictionary of (item, distance) key, value pairs sorted by
                similarity in descending order.
        """
        _n = n
        if self.model is None:
            raise AttributeError('Must have a fit model in order'
                                 ' to call similarity.')
        if self.index is None:
            print "no index, creating our own"
            print
            from gensim.similarities import MatrixSimilarity
            self.index = MatrixSimilarity(self.corpus)

        # if n is larger than the number of documents, return 1 result
        _n = n * (n < len(self.index)) + (1 - (n < len(self.index))) * 1

        if _n == 1:
            print("{n} too large a number, returning 1 result instead.").format(n=n)
        doc_lsi = self.model[doc]
        sims = self.index[doc_lsi]
        sims = sorted(enumerate(sims), key=lambda item: -item[1])[0:_n]
        results = ["Org: {0} -- {s}".format(data_names[i].encode('utf8').strip(), s=j)
                   for i, j in sims]
        return results




from sklearn.base import BaseEstimator, TransformerMixin


class TextProcessor(BaseEstimator, TransformerMixin):
    """
    Custom sklearn transformer to preprocess data.
    Data fit to should be raw document (not split by tokens).
    """

    def __init__(self, remove_stopwords=True, do_lemmatize=False, do_stem=False, stop_words=None, return_list=True):
        """
        Parameters
        ----------
        remove_stopwords: Boolean
            Whether or not to remove stopwords.
        do_lemmatize: Boolean
            Whether or not to lemmatize tokens.
        do_stem: Boolean
            Whether or not to stem tokens.
        stop_words: iterable
            List of stop words, each word of type str.
        return_list: Boolean
            Whether or not to return list or raw document.
            Some sklearn vectorizers prefer a raw document.

        Note - if neither `do_lemmatize` nor `do_stem` are specified, the transform
        method will preprocess using `fast_process`, which is much faster.
        """
        from nltk.corpus import stopwords
        from nltk.stem import PorterStemmer, WordNetLemmatizer

        if stop_words:
            self.stopwords = set(stop_words)
        else:
            self.stopwords = set(stopwords.words('english'))
        self.lemmatizer = WordNetLemmatizer()
        self.stemmer = PorterStemmer()
        self.do_lemmatize = do_lemmatize
        self.do_stem = do_stem
        self.remove_stopwords = remove_stopwords
        # if not stemming or lemmatizing, use faster preprocessing
        if do_lemmatize or do_stem:
            self.preprocess = self.process
        else:
            self.preprocess = self.fast_process
        self.return_list = return_list

    def stem(self, token):
        """
        Stems word `token` via nltk's PorterStemmer.
        """
        return self.stemmer.stem(token)

    def lemmatize(self, token, pos_tag):
        """
        Lemmatizes token via nltk's WordNetLemmatizer.
        This method bust be called on a tuple of (token, post_tag).

        Parameters
        ----------
        token : str
            Token to be processed.
        pos_tag : str
            Part-of-Speech tag for `token`.

        Returns
        -------
            str: Lemmatized token.

        """
        from nltk.corpus import wordnet as wn

        tag = {
            'N': wn.NOUN,
            'V': wn.VERB,
            'R': wn.ADV,
            'J': wn.ADJ,
            'S': wn.ADJ_SAT
        }.get(pos_tag[0], wn.NOUN)
        return self.lemmatizer.lemmatize(token, tag)

    def clean_token(self, token):
        """
        Cleans `token` by performing the following steps:
            1. Converts to lower-case and strips whitespace.
            2. If `token` is a stopword and `remove_stopwords` is True,
               return an empty string.
            3. Removes non-alpha characters from word.

        Parameters
        ----------
        token : str
            String to be cleaned.

        Returns
        -------
            c_token (str): Cleaned string.
        """
        c_token = token.lower().strip()
        if self.remove_stopwords and c_token in self.stopwords:
            return ''
        # remove non-alpha characters
        c_token = filter(lambda x: x.isalpha(), c_token)
        return c_token

    def process(self, document):
        """
        Preprocessing method that stems and/or lemmatizes token
        during a single pass through data.
        """
        cleaned_tokens = []
        if self.do_lemmatize:
            from nltk import pos_tag as nltk_pos_tag

            # lemmatizing requires that each token has an associated POS tag
            postag_doc = nltk_pos_tag(document.split())
            for token, tag in postag_doc:
                token = self.clean_token(token)
                token = self.lemmatize(token, tag)
                if self.do_stem:
                    token = self.stem(token)
                cleaned_tokens.append(token)
        else:
            for token in document.split():
                token = self.clean_token(token)
                if self.do_stem:
                    token = self.stem(token)
                cleaned_tokens.append(token)

        keep_tokens = [ctoken for ctoken in cleaned_tokens if ctoken != '']
        # determine whether we should return list of tokens or single str
        if not self.return_list:
            return ' '.join(keep_tokens)
        return keep_tokens

    def fast_process(self, document):
        """
        Fast preprocessing of document without stemming or lemmatizing tokens.
        Suitable preprocessing for word2vec.

        Parameters
        ----------
        document: str
            Raw text of document.

        Returns
        -------
            iterable: List of reprocessed tokens from document.
        """
        doc = [
            self.clean_token(token)
            for token in document.split()
        ]
        clean_doc = [token for token in doc if token != '']
        if not self.return_list:
            return ' '.join(clean_doc)
        return clean_doc

    def fit(self, X, y=None):
        """
        `fit` method required for inclusion in sklearn pipeline object.
        Returns data as is.
        """
        return self

    def transform(self, documents):
        """
        Preprocesses each document in `documents`.

        Parameters
        ----------
        documents: iterable
            List of documents. Each document raw, untokenized text.

        Returns
        -------
            iterable: list of lists. Each list contains the preprocessed
            tokens for a document.
        """
        document_list = []
        for document in documents:
            document_list.append(self.preprocess(document))
        return document_list




import json
import requests

import pandas as pd
from bs4 import BeautifulSoup

def load_json(file_path):
    with open(file_path) as json_data:
        d = json.load(json_data)
    return d

def fetch_url(url, tags, verify, strip_unicode):
        """
        Scrapes and fetches front-page text content for a given url.
        **Note** - The scraping methods in this are particular to the websites
        provided by the irs 990 forms. In particular, the conditional statements
        dealing with the "www." and "http" stuff may not work across the board for
        different websites. The 990 websites being scraped are for non-profits, and
        often don't have the most secure sites. Thus this seemingly ad-hoc code works
        well.

        Parameters
        ----------
        url : str
            URL of website to be scraped.
        tags : iterable
            List of html tags in-which to grab content.
        verify : Boolean (default=True)
            Whether or not to ignore verification of SSL certificate.
        strip_unicode : Boolean (default=True)
            Whether or not to strip unicode values from result.

        Returns
        -------
            iterable : List of unicode-stripped tokens from front-page of `url`.
        """
        from bs4 import BeautifulSoup
        orig_url = url
        url = url.lower()
        if "www." in url:
            url = url.split("www.")[-1].split("/")[0]
        elif "http://" in url:
            url = url.strip("http://")
        domain = "http://" + url
        html = requests.get(domain, verify=verify, headers=self.headers)
        soup = BeautifulSoup(html.content, "html.parser")
        site_text = []
        for tag in soup.find_all(tags):
            site_text.append(tag.get_text().strip())
        all_text = " ".join(site_text)
        cleaned_text = re.sub('\s+', ' ', all_text)
        if strip_unicode:
            cleaned_text = ''.join([i if ord(i) < 128 else ' ' for i in cleaned_text])
        return pd.DataFrame({"url": [orig_url], "url_text": [cleaned_text]})





name_list = []



"""
Implementations algorithms for vectorizing text, all inheriting from sklearn BaseEstimators
and TransformerMixin.

This allows for use in pipeline objects, as well as provides a familiar API for preprocessing.

Example Code
-------------
>> from classypy.ml.transformers import TextProcessor
>> from classypy.ml.pipeline import process_and_vectorize_pipeline
>> tfidf = process_and_vectorize_pipeline(TextProcessor(), GensimTfid(use_sparse_representation=True))
>> tfidf.fit(data)
>> tfidf_data = list(tfidf.transform(data))  # call list because return is generator
"""

import os

from sklearn.base import BaseEstimator, TransformerMixin


class GensimBOW(BaseEstimator, TransformerMixin):
    """
    Custom sklearn transformer to convert tokenized,
    preprocessed data to bag-of-words representation.
    """

    def __init__(self, id2word_path=None, use_sparse_representation=False):
        """
        Parameters
        ----------
        id2word_path : str
            Path to location of gensim id2word dict.
            If specified, the model will load and use this object
            as its id2_word dict.
        use_sparse_representation: Boolean (default=False)
            When True, a sparse representation of the array is returned.
                Use this when feeding into a gensim model.
            When False, the full array is returned.
                Use this if feeding into sklearn estimator.
        """
        self.id2word = None
        self.use_sparse_representation = use_sparse_representation
        if id2word_path:
            self._load(id2word_path=id2word_path)

    def _load(self, id2word_path):
        """
        If self.id2word_path specified, loads gensim.id2word dict from path.

        Parameters
        ----------
        id2word_path: str
            File-path designating where self.id2word should be saved.
        """
        from gensim.corpora.dictionary import Dictionary
        if not os.path.exists(id2word_path):
            raise IOError('The provided file path to id2word_path was not found.'
                          'Please ensure that the argument is the correct path.')
        self.id2word = Dictionary().load(id2word_path)

    def save(self, id2word_path):
        """
        Saves self.id2word to id2word_path.
        If id2word does not exist, AttributeError is raised.

        Parameters
        ----------
        id2word_path: str
            File-path designating where self.id2word should be saved.
        """
        if not self.id2word:
            raise AttributeError('Nothing to save yet, please run .fit first.')
        self.id2word.save(id2word_path)

    def fit(self, documents, labels=None):
        """
        Creates map between words and their integer ids,
        storing it as `self.id2word`.

        Parameters
        ----------
        documents: iterable
            List of documents; each document a list of preprocessed tokens.
        labels:
            Optional list of same size as documents, specifying label for each document.
        """
        from gensim.corpora.dictionary import Dictionary
        self.id2word = Dictionary(documents)

    def transform(self, documents):
        """
        Converts a collection of words to its bag-of-words representation.

        Parameters
        ----------
        documents: iterable
            List of documents. Each document must be a list of tokens.

        Returns
        -------
            generator: yields vectorized representation of each document.
        """
        from gensim.matutils import sparse2full
        if self.id2word is None:
            raise AttributeError('Must have a fit id2word in order'
                                 ' to call transform.')

        def generator():
            """
            Closure to mutate return type depending on value of `use_sparse_representation`.
            """
            for document in documents:
                docbow = self.id2word.doc2bow(document)
                if self.use_sparse_representation:
                    yield docbow
                else:
                    yield sparse2full(docbow, len(self.id2word))
        return list(generator())


class GensimTfidf(BaseEstimator, TransformerMixin):
    """
    Custom sklearn transformer to convert tokenized,
    preprocessed data to tf-idf representation.
    """
    def __init__(self, tfidf_path=None, dictionary_path=None, use_sparse_representation=False):
        """
        Instantiate GensimTfidf object. If loading previously fit Dictionary and
        TfidfModel, you must specify a path to both the Dictionary and the TfidfModel.

        Parameters
        ----------
        tfidf_path : str
            Path to location of saved gensim TfidfModel.
            If specified, the model will load and use this object
            as its TfidfModel.
        dictionary_path : str
            Path to location of saved gensim Dictionary.
            If specified, the model will load and use this object
            as its Dictionary.
        use_sparse_representation: Boolean (default=False)
            When True, a sparse representation of the array is returned.
                Use this when feeding into a gensim model.
            When False, the full array is returned.
                Use this if feeding into sklearn estimator.
        """
        self.use_sparse_representation = use_sparse_representation
        self.dictionary = None
        self.tfidf = None
        # if both paths specified, load object
        if tfidf_path and dictionary_path:
            self._load(tfidf_path=tfidf_path, dictionary_path=dictionary_path)
        elif tfidf_path or dictionary_path:
            raise AttributeError('If loading pre-fit Dictionary and TfidfModel,'
                                 ' both must be specified, not just one.')

    def _load(self, tfidf_path, dictionary_path):
        """
        If specified, attempts to load gensim TfidfModel from `tfidf_path`
        and gensim Dictionary from `dictionary_path`.

        Parameters
        ----------
        tfidf_path: str
            File-path designating where self.tfidf should be saved.
        dictionary_path: str
            File-path designating where self.dictionary should be saved.
        """
        from gensim.models import TfidfModel
        from gensim.corpora.dictionary import Dictionary
        if not os.path.exists(tfidf_path):
            raise IOError('The provided file path to the TfidfModel was not found.'
                          'Please ensure that the argument is the correct path.')
        if not os.path.exists(dictionary_path):
            raise IOError('The provided file path to the Dictionary was not found.'
                          'Please ensure that the argument is the correct path.')
        self.tfidf = TfidfModel().load(tfidf_path)
        self.dictionary = Dictionary().load(dictionary_path)

    def save(self, tfidf_path, dictionary_path):
        """
        Saves objects from fit process: gensim.TfidfModel to `tfidf_path`
        and gensim.Dictionary to `dictionary_path`.
        If either self.tfidf or self.dictionary does not exist, an
        AttributeError is raised.

        Parameters
        ----------
        tfidf_path: str
            File-path designating where self.tfidf should be saved.
        dictionary_path: str
            File-path designating where self.dictionary should be saved.
        """
        if not (self.tfidf and self.dictionary):
            raise AttributeError('Nothing to save yet, please run .fit first.')
        self.tfidf.save(tfidf_path)
        self.dictionary.save(dictionary_path)

    def fit(self, documents, labels=None):
        """
        Fits a gensim TfidfModel to documents.

        Parameters
        ----------
        documents: iterable
            List of documents. Each document must be a list of preprocessed tokens.
        labels: iterable
            Optional list of same size as documents, specifying label for each document.

        """
        from gensim.models import TfidfModel
        from gensim.corpora.dictionary import Dictionary
        self.dictionary = Dictionary(documents)
        self.tfidf = TfidfModel(
            [
                self.dictionary.doc2bow(doc) for doc in documents
            ], id2word=self.dictionary)
        return self

    def transform(self, documents):
        """
        Returns a vectorized embedding of each document in documents.

        Parameters
        -----------
        documents: iterable
            List of documents. Each document must be a list of tokens.

        Returns
        -------
            iterable: list of vectorized documents.
        """
        from gensim.matutils import sparse2full
        if self.dictionary is None:
            raise AttributeError('Must have a fit vocab in order'
                                 ' to call transform.')

        def generator():
            """
            Closure to mutate return type depending on value of `use_sparse_representation`.
            """
            for document in documents:
                vec = self.tfidf[self.dictionary.doc2bow(document)]
                if self.use_sparse_representation:
                    yield vec
                else:
                    yield sparse2full(vec, len(self.dictionary))
        return list(generator())




from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='artcamp',
      version='0.158',
      description='Query similarity between articles, orgs, and/or campaigns',
      url='http://github.com/jwilber/artcamp',
      author='Jared Wilber',
      author_email='jwilber@classy.org',
      package_data={'artcamp': 'data/*'},
      include_package_data=True,
      license='MIT',
      packages=['artcamp'],
      install_requires=[
          'gensim',
          'scikit-learn',
      ],
      zip_safe=False)



from .lsi import GensimLsi
from .cleaner import TextProcessor
from .vectorizers import GensimTfidf
from .model import OrgSim


__all__ = ['GensimLsi',
           'TextProcessor',
           'GensimTfidf',
           'OrgSim']
